#############################################################################
##
#W  interfac.gi          GAP 4 package AtlasRep                 Thomas Breuer
##
#H  @(#)$Id: interfac.gi,v 1.51 2008/06/16 17:34:37 gap Exp $
##
#Y  Copyright (C)  2001,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
##
##  This file contains the implementation part of the ``high level'' {\GAP}
##  interface to the {\ATLAS} of Group Representations.
##
Revision.( "atlasrep/gap/interfac_gi" ) :=
    "@(#)$Id: interfac.gi,v 1.51 2008/06/16 17:34:37 gap Exp $";


#############################################################################
##
#F  DisplayAtlasInfoOverview( <gapnames>, <tocs> )
##
InstallGlobalFunction( DisplayAtlasInfoOverview, function( gapnames, tocs )
    local result, columns, widths, width, i, j;

    # Consider only those names for which actually information is available.
    # (The ordering shall be the same as in the input.)
    if gapnames = "all" then
      gapnames:= AtlasOfGroupRepresentationsInfo.GAPnames;
    else
      gapnames:= Filtered( List( gapnames,
                     i -> First( AtlasOfGroupRepresentationsInfo.GAPnames,
                                 pair -> pair[1] = i ) ),
                           x -> x <> fail );
    fi;
    gapnames:= Filtered( gapnames,
                   x -> ForAny( tocs, toc -> IsBound( toc.( x[2] ) ) ) );
    if IsEmpty( gapnames ) then
      return;
    fi;

    result:= "";

    # Compute the data of the columns.
    columns:= [ [ "group", "l", List( gapnames, x -> [ x[1], false ] ) ] ];
    Append( columns, List( AGRDataTypes( "rep", "prg" ),
                    type ->
           [ type[2].DisplayOverviewInfo[1],
             type[2].DisplayOverviewInfo[2],
             List( gapnames,
                   n -> type[2].DisplayOverviewInfo[3]( tocs, n[2] ) ) ] ) );

    # Compute the appropriate column widths.
    widths:= [];
    for i in columns do
      width:= Maximum( Length( i[1] ),       # the header string shall fit
                  Maximum( List( i[3], y -> Length( y[1] ) ) ) );
      if width <> 0 then
        width:= width + 2;
      fi;
      if i[2] = "l" then
        width:= -width;
      fi;
      Add( widths, width );
    od;

    # Print the header line.
    for j in [ 1 .. Length( columns ) ] do
      Append( result, FormattedString( columns[j][1], widths[j] ) );
    od;
    Append( result, "\n" );
    Append( result, ListWithIdenticalEntries( Sum( widths, AbsInt ), '-' ) );
    Append( result, "\n" );

    # Print the information for each group.
    for i in [ 1 .. Length( gapnames ) ] do
      if ForAny( columns, x -> x[3][i][2] ) then
        columns[1][3][i][1]:= Concatenation( columns[1][3][i][1],
            AtlasOfGroupRepresentationsInfo.markprivate );
      fi;
      for j in [ 1 .. Length( columns ) ] do
        Append( result, FormattedString( columns[j][3][i][1], widths[j] ) );
      od;
      Append( result, "\n" );
    od;

    AtlasOfGroupRepresentationsInfo.displayFunction( result );
    end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsInfoPRG( <gapname>, <tocs>, <name>, <std> )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsInfoPRG,
    function( gapname, tocs, name, std )
    local stdavail, j, toc, record, type, list, header;

    # If the standardization is prescribed then do not mention it.
    # Otherwise if all information refers to the same standardization then
    # print just one line.
    # Otherwise print the standardization for each entry.
    stdavail:= [];
    if std = true or 1 < Length( std ) then
      for toc in tocs do
        if IsBound( toc.( name ) ) then
          record:= toc.( name );
          for type in AGRDataTypes( "prg" ) do
            if IsBound( record.( type[1] ) ) then
              for list in record.( type[1] ) do
                if std = true or list[1] in std then
                  AddSet( stdavail, list[1] );
                fi;
              od;
            fi;
          od;
        fi;
      od;
    fi;

    # Create the header line.
    # (Because of `AtlasRepCreateHTMLInfoForGroup',
    # `gapname' must occur as an entry of its own .)
    header:= [ "Programs for G = ", gapname, ":" ];
    if Length( stdavail ) = 1 then
      Append( header, [ "    (all refer to std. generators ",
                        String( stdavail[1] ), ")" ] );
    fi;

    # Collect the info lines for the scripts.
    list:= [];
    for type in AGRDataTypes( "prg" ) do
      Append( list, type[2].DisplayPRG( tocs, name, std, stdavail ) );
    od;

    return rec( header := header,
                list   := list );
    end );


#############################################################################
##
#F  AGREvaluateMinimalityCondition( <gapname>, <conditions> )
##
##  Evaluate conditions involving `"minimal"'.
##  The return value is a copy of <conditions> in which one occurrence of
##  `"minimal"' is replaced by the corresponding number if this is known,
##  or a string describing an info message otherwise.
##
BindGlobal( "AGREvaluateMinimalityCondition", function( gapname, conditions )
    local pos, info, pos2;

    pos:= Position( conditions, "minimal" );
    if pos <> fail and pos <> 1 then
      if   IsIdenticalObj( conditions[ pos-1 ], NrMovedPoints ) then
        # ..., NrMovedPoints, "minimal", ...
        info:= MinimalRepresentationInfo( gapname, NrMovedPoints );
        if info = fail then
          return Concatenation( 
                     "minimal perm. repr. of `", gapname, "' not known" );
        fi;
        conditions:= ShallowCopy( conditions );
        conditions[ pos ]:= info.value;
      elif IsIdenticalObj( conditions[ pos-1 ], Dimension ) then
        pos2:= Position( conditions, Characteristic );
        if pos2 <> fail and pos2 < Length( conditions ) then
          # ..., Characteristic, <p>, ..., Dimension, "minimal", ...
          info:= MinimalRepresentationInfo( gapname,
                     Characteristic, conditions[ pos2+1 ] );
          if info = fail then
            return Concatenation( "minimal matrix repr. of `", gapname,
                                  "' in characteristic ",
                                  conditions[ pos2+1 ], " not known" );
          fi;
          conditions:= ShallowCopy( conditions );
          conditions[ pos ]:= info.value;
        else
          pos2:= Position( conditions, Ring );
          if pos2 <> fail and pos2 < Length( conditions )
                         and IsField( conditions[ pos2+1 ] )
                         and IsFinite( conditions[ pos2+1 ] ) then
            # ..., Ring, <R>, ..., Dimension, "minimal", ...
            info:= MinimalRepresentationInfo( gapname,
                       Size, Size( conditions[ pos2+1 ] ) );
            if info = fail then
              return Concatenation( "minimal matrix repr. of `", gapname,
                                    "' over `", conditions[ pos2+1 ],
                                    "' not known" );
            fi;
            conditions:= ShallowCopy( conditions );
            conditions[ pos ]:= info.value;
          fi;
        fi;
      fi;
    fi;

    return conditions;
    end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsInfoGroup( <gapname>, <tocs>, <name>, <std>,
#F      <conditions> )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsInfoGroup,
    function( gapname, tocs, name, std, conditions )
    local stdavail, nam, toc, record, list, header, i, j, nr, type, entry;

    # If the standardization is prescribed then do not mention it.
    # Otherwise if all information refers to the same standardization then
    # print just one line.
    # Otherwise print the standardization for each entry.
    stdavail:= [];
    if std = true or 1 < Length( std ) then
      for nam in AGRDataTypes( "rep" ) do
        for toc in tocs do
          if IsBound( toc.( name ) ) then
            record:= toc.( name );
            if IsBound( record.( nam[1] ) ) then
              for list in record.( nam[1] ) do
                if std = true or list[1] in std then
                  AddSet( stdavail, list[1] );
                fi;
              od;
            fi;
          fi;
        od;
      od;
    fi;

    # Construct the header line.
    # (Because of `AtlasRepCreateHTMLInfoForGroup',
    # `gapname' must occur as an entry of its own .)
    header:= [ "Representations for G = ", gapname, ":" ];
    if Length( stdavail ) = 1 then
      Add( header, Concatenation(
           "    (all refer to std. generators ", String( stdavail[1] ),
           ")" ) );
    fi;

    # Evaluate conditions involving `"minimal"'.
    conditions:= AGREvaluateMinimalityCondition( gapname, conditions );
    if IsString( conditions ) and not IsEmpty( conditions ) then
      Info( InfoAtlasRep, 1, conditions );
    fi;
    
    # Collect the information for the given group.
#T sort across the tocs!
    list:= [];
    nr:= 0;
    for type in AGRDataTypes( "rep" ) do
      for j in [ 1 .. Length( tocs ) ] do
        toc:= tocs[j];
        if IsBound( toc.( name ) ) then
          record:= toc.( name );
          if IsBound( record.( type[1] ) ) then
            for i in record.( type[1] ) do
              nr:= nr + 1;
              if std = true or i[1] in std then
                if type[2].AccessGroupCondition( i,
                       ShallowCopy( conditions ) ) then
                  entry:= type[2].DisplayGroup( i );
                  if 1 < j then
                    Append( entry,
                        AtlasOfGroupRepresentationsInfo.markprivate );
                  fi;
                  if 1 < Length( stdavail ) then
                    Append( entry, Concatenation(
                        ",   w.r.t. std. generators ", String( i[1] ) ) );
                  fi;
                  Add( list, [ String( nr ), entry ] );
                fi;
              fi;
            od;
          fi;
        fi;
      od;
    od;

    return rec( header := header,
                list   := list );
    end );


#############################################################################
##
#F  DisplayAtlasInfo(  )
#F  DisplayAtlasInfo( <listofnames> )
#F  DisplayAtlasInfo( <gapname>[, <std>] )
#F  DisplayAtlasInfo( <gapname>[, <std>], IsPermGroup[, true] )
#F  DisplayAtlasInfo( <gapname>[, <std>], NrMovedPoints, <n> )
#F  DisplayAtlasInfo( <gapname>[, <std>], IsMatrixGroup[, true] )
#F  DisplayAtlasInfo( <gapname>[, <std>][, Characteristic, <p>]
#F                                      [, Dimension, <n>] )
#F  DisplayAtlasInfo( <gapname>[, <std>], IsStraightLineProgram[, true] )
#F  DisplayAtlasInfo( <gapname>[, <std>], IsStraightLineDecision[, true] )
#F  DisplayAtlasInfo( <gapname>[, <std>], IsBBoxProgram[, true] )
##
InstallGlobalFunction( DisplayAtlasInfo, function( arg )
    local tocs, gapname, groupname, result, std, argpos, info1, len, line,
          info2;

    if AtlasOfGroupRepresentationsInfo.remote = true then
      tocs:= [ AtlasTableOfContents( "remote" ).TableOfContents ];
    else
      tocs:= [ AtlasTableOfContents( "local" ).TableOfContents ];
    fi;
    Append( tocs, List( AtlasOfGroupRepresentationsInfo.private,
        x -> AtlasOfGroupRepresentationsInfo.TableOfContents.( x[2] ) ) );

    if   Length( arg ) = 0 then
      DisplayAtlasInfoOverview( "all", tocs );
      return;
    elif Length( arg ) = 1 and IsList( arg[1] )
                           and ForAll( arg[1], IsString ) then
      DisplayAtlasInfoOverview( arg[1], tocs );
      return;
    elif not IsString( arg[1] ) then
      Error( "usage: DisplayAtlasInfo( [<listofnames>] ),\n",
             "DisplayAtlasInfo( <gapname>[,<std>],<conditions> )\n" );
    fi;

    gapname:= arg[1];
    groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                       pair -> pair[1] = gapname );
    if groupname = fail then
      Info( InfoAtlasRep, 1,
            "DisplayAtlasInfo: no group with GAP name `", gapname, "'" );
      return;
    fi;

    result:= "";

    if Length( arg ) = 1 or not ( IsInt( arg[2] ) or IsList( arg[2] ) ) then
      std:= true;
      argpos:= 2;
    else
      std:= arg[2];
      if IsInt( std ) then
        std:= [ std ];
      fi;
      argpos:= 3;
    fi;

    if Length( arg ) <> argpos or arg[ argpos ] <> IsStraightLineProgram then

      # `DisplayAtlasInfo( <gapname>[, <std>][, <conditions>] )'
      info1:= AtlasOfGroupRepresentationsInfoGroup( gapname, tocs,
                  groupname[2], std, arg{ [ argpos .. Length( arg ) ] } );
      if not IsEmpty( info1.list ) then
        Append( result, Concatenation(
                  Concatenation( info1.header ), "\n",
                  ListWithIdenticalEntries( Sum( List(
                      info1.header{ [ 1 .. 3 ] }, Length ) ), '-' ),
                  "\n" ) );
        len:= Maximum( List( info1.list, x -> Length( x[1] ) ) );
        for line in info1.list do
          Append( result, Concatenation(
                    FormattedString( line[1], len ), ": ", line[2], "\n" ) );
        od;
      fi;
    fi;
    if ( Length( arg ) = argpos and arg[ argpos ] = IsStraightLineProgram )
       or ( Length( arg ) = argpos + 1
            and arg[ argpos ] = IsStraightLineProgram
            and arg[ argpos + 1 ] = true )
       or Length( arg ) < argpos then

      # `DisplayAtlasInfo( <gapname>[, <std>][, IsStraightLineProgram] )'
      info2:= AtlasOfGroupRepresentationsInfoPRG( gapname, tocs,
                  groupname[2], std );
      if not IsEmpty( info2.list ) then
        if IsBound( info1 ) and not IsEmpty( info1.list ) then
          Append( result, "\n" );
        fi;
        Append( result, Concatenation(
                  Concatenation( info2.header ), "\n",
                  ListWithIdenticalEntries( Sum( List(
                      info2.header{ [ 1 .. 3 ] }, Length ) ), '-' ),
                  "\n",
                  Concatenation( info2.list ) ) );
      fi;

    fi;

    AtlasOfGroupRepresentationsInfo.displayFunction( result );
    end );


#############################################################################
##
#F  AtlasGenerators( <gapname>, <repnr> )
#F  AtlasGenerators( <gapname>, <repnr>, <maxnr> )
#F  AtlasGenerators( <identifier> )
##
##  <identifier> is a list containing at the first position the string
##  <gapname>,
##  at the second position a string or a list of strings
##  (describing filenames),
##  at the third position a positive integer denoting the standardization of
##  the representation,
##  at the fourth position a positive integer describing the common ring of
##  the generators,
##  and at the fifth position, if bound, a positive integer denoting the
##  number of the maximal subgroup to which the representation is restricted.
##
InstallGlobalFunction( AtlasGenerators, function( arg )
    local tocs, identifier, gapname, prefix, groupname, maxnr, file, repnr,
          res, type, j, toc, record, pos, try, gens, name, gen, result, prog;

    if AtlasOfGroupRepresentationsInfo.remote = true then
      tocs:= [ AtlasTableOfContents( "remote" ).TableOfContents ];
    else
      tocs:= [ AtlasTableOfContents( "local" ).TableOfContents ];
    fi;
    Append( tocs, List( AtlasOfGroupRepresentationsInfo.private,
        x -> AtlasOfGroupRepresentationsInfo.TableOfContents.( x[2] ) ) );

    if Length( arg ) = 1 then

      # `AtlasGenerators( <identifier> )'
      identifier:= arg[1];
      if IsRecord( identifier ) and IsBound( identifier.identifier ) then
        identifier:= identifier.identifier;
      fi;
      gapname:= identifier[1];
      if Length( gapname ) = 2 and IsString( gapname[1] ) then
        # file in a private directory
        prefix:= gapname[1];
        gapname:= gapname[2];
      else
        prefix:= "datagens";
      fi;
      groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                         pair -> pair[1] = gapname );
      if IsBound( identifier[5] ) then
        maxnr:= identifier[5];
      fi;
      file:= identifier[2];
      if not IsString( file ) then
        file:= file[1];
      fi;

      # Compute the type, and the current number of the representation.
      repnr:= 0;
      res:= false;
      for type in AGRDataTypes( "rep" ) do
        for j in [ 1 .. Length( tocs ) ] do
          toc:= tocs[j];
          if IsBound( toc.( groupname[2] ) ) then
            record:= toc.( groupname[2] );
            if IsBound( record.( type[1] ) ) then
              pos:= PositionProperty( record.( type[1] ),
                        entry -> entry[ Length( entry ) ] = identifier[2] );
              if pos = fail then
                repnr:= repnr + Length( record.( type[1] ) );
              else
                repnr:= repnr + pos;
                res:= true;
                break;
              fi;
            fi;
          fi;
        od;
        if res then
          break;
        fi;
      od;
      if not res then
        return fail;
      fi;

    elif  ( Length( arg ) = 2 and IsString( arg[1] ) and IsPosInt( arg[2] ) )
       or ( Length( arg ) = 3 and IsString( arg[1] ) and IsPosInt( arg[2] )
                              and IsPosInt( arg[3] ) ) then

      # `AtlasGenerators( <gapname>, <repnr>[, <maxnr>] )'
      gapname:= arg[1];
      groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                         pair -> pair[1] = gapname );
      if groupname = fail then
        Info( InfoAtlasRep, 1,
              "AtlasGenerators: no group with GAP name `", gapname, "'" );
        return fail;
      fi;

      try:= function( repnr, type )
        local j, toc, record;
        for j in [ 1 .. Length( tocs ) ] do
          toc:= tocs[j];
          if IsBound( toc.( groupname[2] ) ) then
            record:= toc.( groupname[2] );
            if IsBound( record.( type ) ) then
              if repnr <= Length( record.( type ) ) then
                return [ j, record.( type )[ repnr ] ];
              fi;
              repnr:= repnr - Length( record.( type ) );
            fi;
          fi;
        od;
        return repnr;
      end;

      repnr:= arg[2];
      res:= repnr;
      for type in AGRDataTypes( "rep" ) do
        res:= try( res, type[1] );
        if not IsInt( res ) then
          break;
        fi;
      od;
      if IsInt( res ) then
        return fail;
      fi;

      if res[1] = 1 then
        prefix:= "datagens";
      else
        prefix:= AtlasOfGroupRepresentationsInfo.private[ res[1]-1 ][2];
      fi;
      res:= res[2];
      identifier:= [ gapname, res[ Length( res) ], res[1], res[2] ];
      if prefix <> "datagens" then
        identifier[1]:= [ prefix, gapname ];
      fi;
      if IsBound( arg[3] ) then
        maxnr:= arg[3];
        identifier[5]:= maxnr;
      fi;

    else
      Error( "usage: AtlasGenerators( <gapname>,<repnr>[,<maxnr>] ) or\n",
             "       AtlasGenerators( <identifier> )" );
    fi;

    # Access the data file(s).
    gens:= AGRFileContents( prefix, groupname[2], identifier[2], type );
    if gens = fail then
      return fail;
    fi;
    result:= rec( generators      := gens,
                  groupname       := gapname,
                  standardization := identifier[3],
                  repnr           := repnr,
                  identifier      := identifier );
    type[2].AddDescribingComponents( result, type );

    if IsBound( maxnr ) then

      Unbind( result.groupname );

      # Compute the straight line program for the restriction
      # (w.r.t. the correct standardization).
      prog:= AtlasProgram( gapname, identifier[3], maxnr );
      if prog = fail then
        return fail;
      fi;

      # Evaluate the straight line program.
      result.generators:= ResultOfStraightLineProgram( prog.program, gens );

      if IsBound( groupname[4] ) and IsBound( groupname[4][ maxnr ] ) then
        result.size:= groupname[4][ maxnr ];
      fi;

    elif IsBound( groupname[3] ) then
      result.size:= groupname[3];
    fi;

    # Return the result.
    return Immutable( result );
    end );


#############################################################################
##
#F  AtlasGeneratingSetInfo( <conditions>, "one" )
#F  AtlasGeneratingSetInfo( <conditions>, "all" )
##
##  This function does the work for `OneAtlasGeneratingSetInfo' and
##  `AllAtlasGeneratingSetInfos'.
##  The first entry in <conditions> can be a group name
##  or a list of group names.
##
BindGlobal( "AtlasGeneratingSetInfo", function( conditions, mode )
    local gapnames, std, tocs, result, gapname, groupname, cond, repnr,
          types, type, j, toc, record, i, id, oneresult;

    # The first argument (if there is one) is a group name,
    # a list of group names, an integer (denoting a standardization),
    # or a function (denoting the first condition).
    if Length( conditions ) = 0 or IsInt( conditions[1] )
                                or IsFunction( conditions[1] ) then
      # The group is not restricted.
      gapnames:= List( AtlasOfGroupRepresentationsInfo.GAPnames,
                       pair -> pair[1] );
    elif IsString( conditions[1] ) then
      # Only one group is considered.
      gapnames:= [ conditions[1] ];
      conditions:= conditions{ [ 2 .. Length( conditions ) ] };
    elif IsList( conditions[1] ) and ForAll( conditions[1], IsString ) then
      # A list of group names is prescribed.
      gapnames:= conditions[1];
      conditions:= conditions{ [ 2 .. Length( conditions ) ] };
    else
      Error( "invalid first argument ", conditions[1] );
    fi;

    if 1 <= Length( conditions ) and
       ( IsPosInt( conditions[1] ) or IsList( conditions[1] ) ) then
      std:= conditions[1];
      if IsPosInt( std ) then
        std:= [ std ];
      fi;
      conditions:= conditions{ [ 2 .. Length( conditions ) ] };
    else
      std:= true;
    fi;

    if AtlasOfGroupRepresentationsInfo.remote = true then
      tocs:= [ AtlasTableOfContents( "remote" ).TableOfContents ];
    else
      tocs:= [ AtlasTableOfContents( "local" ).TableOfContents ];
    fi;
    Append( tocs, List( AtlasOfGroupRepresentationsInfo.private,
        x -> AtlasOfGroupRepresentationsInfo.TableOfContents.( x[2] ) ) );

    result:= [];

    for gapname in gapnames do
      groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                         pair -> pair[1] = gapname );
      if groupname <> fail then

        # Evaluate conditions involving `"minimal"'.
        cond:= AGREvaluateMinimalityCondition( gapname, conditions );
        if IsString( cond ) and not IsEmpty( cond ) then
          Info( InfoAtlasRep, 1, cond ); 
        else

          # Loop over the relevant representations.
          repnr:= 1;
          for type in AGRDataTypes( "rep" ) do
            for j in [ 1 .. Length( tocs ) ] do
              toc:= tocs[j];
              if IsBound( toc.( groupname[2] ) ) then
                record:= toc.( groupname[2] );
                if IsBound( record.( type[1] ) ) then
                  for i in record.( type[1] ) do
                    if     ( std = true or i[1] in std )
                       and type[2].AccessGroupCondition( i,
                               ShallowCopy( cond ) ) then
                      if j = 1 then
                        id:= [ gapname, i[ Length(i) ], i[1], i[2] ];
                      else
                        id:= AtlasOfGroupRepresentationsInfo.private[ j-1 ][2];
                        id:= [ [ id, gapname ], i[ Length(i) ], i[1], i[2] ];
                      fi;
                      oneresult:= rec(
                                       groupname       := gapname,
                                       standardization := id[3],
                                       repnr           := repnr,
                                       identifier      := id,
                                     );
                      type[2].AddDescribingComponents( oneresult, type );
                      if IsBound( groupname[3] ) then
                        oneresult.size:= groupname[3];
                      fi;
                      if mode = "one" then
                        return oneresult;
                      else
                        Add( result, oneresult );
                      fi;
                    fi;
                    repnr:= repnr + 1;
                  od;
                fi;
              fi;
            od;
          od;

        fi;

      fi;
    od;

    # We have checked all available representations.
    if mode = "one" then
      return fail;
    else
      return result;
    fi;
    end );


#############################################################################
##
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>] )
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>], IsPermGroup[, true] )
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>], NrMovedPoints, <n> )
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>], IsMatrixGroup[, true] )
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>][, Characteristic, <p>]
#F                                                 [, Dimension, <m>] )
#F  OneAtlasGeneratingSetInfo( [<gapname>][, <std>][, Ring, <R>]
#F                                                 [, Dimension, <m>] )
##
InstallGlobalFunction( OneAtlasGeneratingSetInfo, function( arg )
    return AtlasGeneratingSetInfo( arg, "one" );
    end );


#############################################################################
##
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>] )
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>], IsPermGroup[, true] )
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>], NrMovedPoints, <n> )
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>], IsMatrixGroup[, true] )
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>][, Characteristic, <p>]
#F                                                  [, Dimension, <m>] )
#F  AllAtlasGeneratingSetInfos( [<gapname>][, <std>][, Ring, <R>]
#F                                                  [, Dimension, <m>] )
##
InstallGlobalFunction( AllAtlasGeneratingSetInfos, function( arg )
    return AtlasGeneratingSetInfo( arg, "all" );
    end );


#############################################################################
##
#F  AtlasGroup( [<gapname>][, <std>] ) 
#F  AtlasGroup( [<gapname>][, <std>], IsPermGroup[, true] )
#F  AtlasGroup( [<gapname>][, <std>], NrMovedPoints, <n> )
#F  AtlasGroup( [<gapname>][, <std>], IsMatrixGroup[, true] )
#F  AtlasGroup( [<gapname>][, <std>][, Characteristic, <p>]
#F                                  [, Dimension, <m>] )
#F  AtlasGroup( [<gapname>][, <std>][, Ring, <R>][, Dimension, <m>] ) 
#F  AtlasGroup( <identifier> )
##  
InstallGlobalFunction( AtlasGroup, function( arg )
    local info, result;

    if Length( arg ) = 1 and IsRecord( arg[1] ) then
      info:= arg[1];
    else
      info:= CallFuncList( OneAtlasGeneratingSetInfo, arg );
    fi;
    if info <> fail then
      info:= AtlasGenerators( info.identifier );
      result:= GroupWithGenerators( info.generators );
      if IsBound( info.size ) then
        SetSize( result, info.size );
      fi;
      return result;
    fi;
    return fail;
    end );


#############################################################################
##
#F  AtlasSubgroup( <gapname>[, <std>], <maxnr> )
#F  AtlasSubgroup( <gapname>[, <std>], IsPermGroup[, true], <maxnr> )
#F  AtlasSubgroup( <gapname>[, <std>], NrMovedPoints, <n>, <maxnr> )
#F  AtlasSubgroup( <gapname>[, <std>], IsMatrixGroup[, true], <maxnr> )
#F  AtlasSubgroup( <gapname>[, <std>][, Characteristic, <p>]
#F                                  [, Dimension, <m>], <maxnr> )
#F  AtlasSubgroup( <gapname>[, <std>][, Ring, <R>]
#F                                  [, Dimension, <m>], <maxnr> )
##
InstallGlobalFunction( AtlasSubgroup, function( arg )
    local maxnr, info, std, prog, gens, result;

    if 1 < Length( arg ) then
      maxnr:= arg[ Length( arg ) ];
      info:= CallFuncList( OneAtlasGeneratingSetInfo,
                           arg{ [ 1 .. Length( arg ) - 1 ] } );
      if info <> fail then
        std:= info.standardization;
        prog:= AtlasProgram( arg[1], std, "maxes", maxnr );
        if prog <> fail then
          gens:= AtlasGenerators( info.groupname, info.repnr );
          if gens <> fail then
            result:= ResultOfStraightLineProgram( prog.program,
                                                  gens.generators );
            result:= GroupWithGenerators( result );
            if IsBound( prog.size ) then
              SetSize( result, prog.size );
            fi;
            return result;
          fi;
        fi;
      fi;
    fi;

    return fail;
    end );


#############################################################################
##
#F  AtlasProgram( <gapname>[, <std>][, "maxes"], <maxnr> )
#F  AtlasProgram( <gapname>[, <std>], "classes" )
#F  AtlasProgram( <gapname>[, <std>], "cyclic" )
#F  AtlasProgram( <gapname>[, <std>], "automorphism", <autname> )
#F  AtlasProgram( <gapname>[, <std>], "check" )
#F  AtlasProgram( <gapname>[, <std>], "pres" )
#F  AtlasProgram( <gapname>[, <std>], "find" )
#F  AtlasProgram( <gapname>, <std>, "restandardize", <std2> )
#F  AtlasProgram( <gapname>[, <std>], "other", <descr> )
#F  AtlasProgram( <identifier> )
##
##  <identifier> is a list containing at the first position the string
##  <gapname>,
##  at the second position a string or a list of strings
##  (describing the filenames involved),
##  and at the third position a positive integer denoting the standardization
##  of the program.
##
InstallGlobalFunction( AtlasProgram, function( arg )
    local identifier, gapname, prefix, groupname, type, result, std,
          argpos, conditions, tocs, j, toc, record, id;

    if Length( arg ) = 1 then

      # `AtlasProgram( <identifier> )'
      identifier:= arg[1];
      gapname:= identifier[1];
      if Length( gapname ) = 2 and IsString( gapname[1] ) then
        prefix:= gapname[1];
        gapname:= gapname[2];
      else
        prefix:= "dataword";
      fi;
      groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                         pair -> pair[1] = gapname );
      if groupname = fail then
        return fail;
      fi;
      for type in AGRDataTypes( "prg" ) do
        result:= type[2].AtlasProgram( type, identifier, prefix,
                                       groupname[2] );
        if result <> fail then
          result.groupname:= gapname;
          return Immutable( result );
        fi;
      od;
      return fail;

    fi;

    # Now handle the cases of more than one argument.
    if Length( arg ) = 0 or not IsString( arg[1] ) then
      Error( "the first argument must be the GAP name of a group" );
    fi;
    gapname:= arg[1];
    groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                       pair -> pair[1] = gapname );
    if groupname = fail then
      Info( InfoAtlasRep, 1,
            "AtlasProgram: no group with GAP name `", gapname, "'" );
      return fail;
    fi;

    if IsInt( arg[2] ) and 2 < Length( arg ) then
      std:= [ arg[2] ];
      argpos:= 3;
    else
      std:= true;
      argpos:= 2;
    fi;
    conditions:= arg{ [ argpos .. Length( arg ) ] };

    # `AtlasProgram( <gapname>[, <std>][, "maxes"], <maxnr> )'
    if Length( conditions ) = 1 and IsInt( conditions[1] ) then
      conditions:= [ "maxes", conditions[1] ];
    fi;

    if AtlasOfGroupRepresentationsInfo.remote = true then
      tocs:= [ AtlasTableOfContents( "remote" ).TableOfContents ];
    else
      tocs:= [ AtlasTableOfContents( "local" ).TableOfContents ];
    fi;
    Append( tocs, List( AtlasOfGroupRepresentationsInfo.private,
        x -> AtlasOfGroupRepresentationsInfo.TableOfContents.( x[2] ) ) );

    for j in [ 1 .. Length( tocs ) ] do
      toc:= tocs[j];
      if IsBound( toc.( groupname[2] ) ) then
        record:= toc.( groupname[2] );
        for type in AGRDataTypes( "prg" ) do
          id:= type[2].AccessPRG( record, std, conditions );
          if id <> fail then
            if j = 1 then
              id:= Concatenation( [ groupname[1] ], id );
            else
              id:= Concatenation( [
                [ AtlasOfGroupRepresentationsInfo.private[ j-1 ][2],
                  groupname[1] ] ], id );
            fi;
            return AtlasProgram( id );
          fi;
        od;
      fi;
    od;

    # No program was found.
    Info( InfoAtlasRep, 2,
          "no program for conditions ", conditions, "\n",
          "#I  of the group with GAP name `", groupname[1], "'" );
    return fail;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsShowUserParameters()
##
InstallGlobalFunction( AtlasOfGroupRepresentationsShowUserParameters,
  function()
    local str, prefix, pair, r;

    str:= "access to remote data:    ";
    if AtlasOfGroupRepresentationsInfo.remote then
      Append( str, "yes\n" );
    else
      Append( str, "no\n" );
    fi;

    prefix:= "servers:                  ";
    for pair in AtlasOfGroupRepresentationsInfo.servers do
      Append( str, prefix );
      prefix:= "                          ";
      Append( str, pair[1] );
      Append( str, "/" );
      Append( str, pair[2] );
      Append( str, "\n" );
    od;

    Append( str, "access remote data:       " );
    if   not IsBound( AtlasOfGroupRepresentationsInfo.wget )
         or not AtlasOfGroupRepresentationsInfo.wget in [ true, false ] then
      Append( str, "via the IO package (preferred) or wget\n" );
    elif AtlasOfGroupRepresentationsInfo.wget = true then
      Append( str, "only via wget\n" );
    else
      Append( str, "only via the IO package\n" );
    fi;

    Append( str, "compress data files:      " );
    if AtlasOfGroupRepresentationsInfo.compress then
      Append( str, "yes\n" );
    else
      Append( str, "no\n" );
    fi;

    Append( str, "display overviews via:    " );
    Append( str,
        NameFunction( AtlasOfGroupRepresentationsInfo.displayFunction ) );
    Append( str, "\n" );

    prefix:= "access functions:         ";
    for r in Reversed( AtlasOfGroupRepresentationsInfo.accessFunctions ) do
      Append( str, prefix );
      prefix:= "                          ";
      Append( str, r.description );
      if r.active = true then
        Append( str, " [enabled]\n" );
      else
        Append( str, " [disabled]\n" );
      fi;
    od;

    Append( str, "read MeatAxe text files:  " );
    if IsBound( CMeatAxe.FastRead ) and CMeatAxe.FastRead = true then
      Append( str, "fast\n" );
    else
      Append( str, "minimizing the space\n" );
    fi;

    Print( str );
end );


#############################################################################
##
#E

