#############################################################################
##
#W  drawapery.gi      GAP library     Manuel Delgado <mdelgado@fc.up.pt>
#W                                 Pedro A. Garcia-Sanchez <pedro@ugr.es>
#W                                     Jose Morais    <josejoao@fc.up.pt>
##
#H  @(#)$Id: drawgraph.gi,v 1.12 $
##
#Y  Copyright (C)  2004,  CMUP, Universidade do Porto, Portugal
##
##  The functions in this file make use of the external program dot (from
##  the freely available software package graphviz, for graph visualization)
##  to display the graphs.
##
############################################################################

#############################################################################
##
#F  DrawAperyListOfNumericalSemigroup(ap)
##
##  <ap> is the Apery set of a numerical semigroup.
##  This function draws a graph whose vertices are
##  the elements of <ap> and the arrow u -> v exists
##  iff v - u is in <ap>.
##
#############################################################################
InstallGlobalFunction(DrawAperyListOfNumericalSemigroup, function(ap)
    local i, j, E, ap2, e, path, name, gv, str, dot, tdir,
          d__, name__, s1__;

    if not IsAperyListOfNumericalSemigroup(ap) then
        Error("The argument must be the Apery set of a numerical semigroup");
    fi;

    # Compute the set of edges of the digraph
    ap2 := Set(ap);
    E := [];
    for i in ap do
        for j in ap do
            if j-i in ap2 then
                Add(E, [i,j]);
            fi;
        od;
    od;


    # The file where dot graph will be written
    tdir := DirectoryTemporary();
    if tdir = fail then
        tdir := Directory(GAPInfo.UserHome);
    fi;
    name := Filename(tdir, Concatenation("aperygraph", ".dot"));
    PrintTo(name, "digraph  AperyGraph{\n");


    # Search for a program to display .ps
    path := DirectoriesSystemPrograms();
    gv := Filename( path, "evince" );
    if gv = fail then
        gv := Filename( path, "ggv" );
    fi;
    if gv = fail then
        gv := Filename( path, "gsview32" );
    fi;
    if gv = fail then
        gv := Filename( path, "gsview64" );
    fi;
    if gv = fail then
        gv := Filename( path, "gv" );
    fi;
    if gv = fail then
        Error("Please install evince, ggv, gsview or gv");
    fi;


    if ARCH_IS_UNIX( ) then
        dot := Filename( path, "dot" );
    elif ARCH_IS_WINDOWS( ) then
        for d__ in path do
            name__ := LowercaseString(Filename(d__, ""));
            s1__ := ReplacedString(name__, "graphviz", "xx");
            if s1__ <> name__ then
                dot := Filename( d__, "dot" );
                break;
            fi;
        od;
    else

    fi;
    if dot = fail then
        Error("Please install GraphViz (http://www.graphviz.org )");
    fi;


    # Write the dot file to a string
    str := "";
    for e in E do
        Append(str, Concatenation(String(e[1]), " -> ", String(e[2]), ";\n"));
    od;
    for i in ap2 do
        Append(str, Concatenation(String(i), ";\n"));
    od;
    Append(str, "}\n");


    # Write the string to the file
    AppendTo(name, str);

    # Run dot and display graph
#    Exec(Concatenation(dot, " -Tps ", name, " -o ", name, ".ps"));
#    Exec(Concatenation(gv, " ", name, ".ps 2>/dev/null &"));

    Process(tdir, dot, InputTextUser(), OutputTextUser(), ["-Tps", "-o", "aperygraph.dot.ps", "aperygraph.dot"]);
    Print(Concatenation("Displaying file: ", Filename(tdir, "aperygraph.dot.ps"), "\n"));
    if DrawingsExtraFormat <> "none" then
        Process(tdir, dot, InputTextUser(), OutputTextUser(), [Concatenation("-T", DrawingsExtraFormat), "-o", Concatenation(name, ".", DrawingsExtraFormat), name]);
        Print(Concatenation("The extra output format file: ", name, ".", DrawingsExtraFormat, "\nhas also been created.\n"));
    fi;
    if ARCH_IS_UNIX( ) then
        Exec(Concatenation("cd ", Filename(tdir, ""), "; ", gv, " aperygraph.dot.ps 1>/dev/null 2>/dev/null &"));
    elif ARCH_IS_WINDOWS( ) then
        Process(tdir, gv, InputTextUser(), OutputTextUser(), ["aperygraph.dot.ps"]);
    else

    fi;
end);

#############################################################################
##
#F  XDrawAperyListOfNumericalSemigroup(ap)
##
##  The same as above but to be called by XNumericalSemigroup.
##
#############################################################################
InstallGlobalFunction(XDrawAperyListOfNumericalSemigroup, function(S, n)
    local i, j, E, ap2, e, path, name, gv, str, dot, d__, name__, s1__, tdir, ap;

	ap := XAperyListOfNumericalSemigroupWRTElement(S, n);
	if ap = fail then
        return;
    fi;

    if not IsAperyListOfNumericalSemigroup(ap) then
        Print("The argument must be the Apery set of a numerical semigroup");
        return;
    fi;

    # Compute the set of edges of the digraph
    ap2 := Set(ap);
    E := [];
    for i in ap do
        for j in ap do
            if j-i in ap2 then
                Add(E, [i,j]);
            fi;
        od;
    od;


    # The file where dot graph will be written
    tdir := DirectoryTemporary();
    if tdir = fail then
        tdir := Directory(GAPInfo.UserHome);
    fi;
    name := Filename(tdir, Concatenation("aperygraph", ".dot"));
    PrintTo(name, "digraph  AperyGraph{\n");


    # Search for a program to display .ps
    path := DirectoriesSystemPrograms();
    gv := Filename( path, "evince" );
    if gv = fail then
        gv := Filename( path, "ggv" );
    fi;
    if gv = fail then
        gv := Filename( path, "gsview32" );
    fi;
    if gv = fail then
        gv := Filename( path, "gsview64" );
    fi;
    if gv = fail then
        gv := Filename( path, "gv" );
    fi;
    if gv = fail then
        Print("Please install evince, ggv, gsview or gv");
        return;
    fi;


    if ARCH_IS_UNIX( ) then
        dot := Filename( path, "dot" );
    elif ARCH_IS_WINDOWS( ) then
        for d__ in path do
            name__ := LowercaseString(Filename(d__, ""));
            s1__ := ReplacedString(name__, "graphviz", "xx");
            if s1__ <> name__ then
                dot := Filename( d__, "dot" );
                break;
            fi;
        od;
    else

    fi;
    if dot = fail then
        Print("Please install GraphViz (http://www.graphviz.org )");
        return;
    fi;



    # Write the dot file to a string
    str := "";
    for e in E do
        Append(str, Concatenation(String(e[1]), " -> ", String(e[2]), ";\n"));
    od;
    for i in ap2 do
        Append(str, Concatenation(String(i), ";\n"));
    od;
    Append(str, "}\n");


    # Write the string to the file
    AppendTo(name, str);

    # Run dot and display graph
#    Exec(Concatenation(dot, " -Tps ", name, " -o ", name, ".ps"));
#    Exec(Concatenation(gv, " ", name, ".ps 2>/dev/null &"));

    Process(tdir, dot, InputTextUser(), OutputTextUser(), ["-Tps", "-o", "aperygraph.dot.ps", "aperygraph.dot"]);
    Print(Concatenation("Displaying file: ", Filename(tdir, "aperygraph.dot.ps"), "\n"));
    if DrawingsExtraFormat <> "none" then
        Process(tdir, dot, InputTextUser(), OutputTextUser(), [Concatenation("-T", DrawingsExtraFormat), "-o", Concatenation(name, ".", DrawingsExtraFormat), name]);
        Print(Concatenation("The extra output format file: ", name, ".", DrawingsExtraFormat, "\nhas also been created.\n"));
    fi;
    if ARCH_IS_UNIX( ) then
        Exec(Concatenation("cd ", Filename(tdir, ""), "; ", gv, " aperygraph.dot.ps 1>/dev/null 2>/dev/null &"));
    elif ARCH_IS_WINDOWS( ) then
        Process(tdir, gv, InputTextUser(), OutputTextUser(), ["aperygraph.dot.ps"]);
    else

    fi;

end);

